"""Uncertainty quantification for Landmarker."""

from .uncertainty import (
    MR2C2R,
    MR2CCP,
    ConformalRegressorBonferroni,
    ConformalRegressorMahalanobis,
    ConformalRegressorMaxNonconformity,
    ContourHuggingRegressor,
    MultivariateNormalRegressor,
    resize_landmarks,
    transform_heatmap_to_original_size,
    transform_heatmap_to_original_size_numpy,
)

__all__ = [
    "ConformalRegressorMahalanobis",
    "ConformalRegressorMaxNonconformity",
    "ConformalRegressorBonferroni",
    "MR2CCP",
    "MR2C2R",
    "MultivariateNormalRegressor",
    "ContourHuggingRegressor",
    "transform_heatmap_to_original_size_numpy",
    "transform_heatmap_to_original_size",
    "resize_landmarks",
]



"""Evaluation visualization functions."""

from typing import Optional

import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import seaborn as sns  # type: ignore
import torch

from landmarker.metrics.metrics import point_error, sdr  # type: ignore


def plot_cpe(
    true_landmarks: torch.Tensor,
    pred_landmarks: torch.Tensor,
    dim: tuple[int, ...] | torch.Tensor,
    dim_orig: torch.Tensor,
    pixel_spacing: torch.Tensor,
    padding: torch.Tensor,
    class_names: Optional[list[str]] = None,
    group: bool = True,
    title: Optional[str] = None,
    save_path: Optional[str] = None,
    stat: str = "proportion",
    unit: str = "mm",
    kind: str = "ecdf",
):
    """Calculate the cumulative point-to-point error (CPE) and plot the CPE curve.

    Args:
        y_true : numpy.ndarray
            The true values of the target variable, with shape (n, p, 2).
            n is the number of samples, p is the number of landmarks,
            and 2 is respectively the y and x coordinates of the landmarks.
        y_pred : numpy.ndarray
            The predicted values of the target variable, with shape (n, p, 2).
            n is the number of samples, p is the number of landmarks,
            and 2 is respectively the y and x coordinates of the landmarks.
        class_names: list
            The names of the landmarks. The default is None.
        group : bool, optional
            Whether to group the CPE curves by landmarks. The default is True.
        title : str, optional
            The title of the plot. The default is None.
        save_path : str, optional
            The path to save the plot. The default is None.
        stat : str, optional
            The type of statistic to plot. The default is 'proportion'.
        unit : str, optional
            The unit of distance. The default is 'mm'.
        kind : str, optional
            The type of plot. The default is 'ecdf'.
            Possible values are 'ecdf', 'kde', and 'hist'.
    """
    # Calculate the point-to-point error (PE)
    pe = point_error(
        true_landmarks, pred_landmarks, dim, dim_orig, pixel_spacing, padding, reduction="none"
    )
    pe = pe.cpu().detach().numpy()
    if len(pe.shape) == 3:
        pe = np.nansum(pe, axis=-1)

    # Plot the CPE curve
    if group:
        # Calculate the mean point-to-point error (MPE) for each sample
        pe = np.nanmean(pe, axis=1)
        if kind == "ecdf":
            ax = sns.ecdfplot(x=pe, stat=stat, label="Sample MPE")
        elif kind == "kde":
            ax = sns.kdeplot(x=pe, cumulative=True, label="Sample MPE")
        elif kind == "hist":
            ax = sns.histplot(x=pe, stat=stat, cumulative=True, label="Sample MPE")
        else:
            raise ValueError(f"Invalid kind: {kind!r}.")
    else:
        if class_names is None:
            class_names = [f"Landmark {i}" for i in range(true_landmarks.shape[1])]
        fig, ax = plt.subplots()
        for i, class_name in enumerate(class_names):
            if kind == "ecdf":
                sns.ecdfplot(x=pe[:, i], stat=stat, ax=ax, label=class_name)
            elif kind == "kde":
                sns.kdeplot(x=pe[:, i], cumulative=True, ax=ax, label=class_name)
            elif kind == "hist":
                sns.histplot(x=pe[:, i], stat=stat, cumulative=True, ax=ax, label=class_name)
            else:
                raise ValueError(f"Invalid kind: {kind!r}.")
    ax.set(xlabel=f"IPE in {unit}", ylabel="Proportion of images", title=title)
    ax.grid()
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    if save_path is not None:
        fig.savefig(save_path)
    plt.show()


def detection_report(
    true_landmarks: torch.Tensor,
    pred_landmarks: torch.Tensor,
    dim: tuple[int, ...] | torch.Tensor,
    dim_orig: torch.Tensor,
    pixel_spacing: torch.Tensor,
    padding: torch.Tensor,
    class_names: Optional[list[str]] = None,
    radius: list[float] = [2, 2.5, 3, 4],
    digits: int = 2,
    unit: str = "mm",
    output_dict: bool = False,
):
    """Calculate the detection report.

    Args:
        true_landmarks : torch.Tensor
            The true landmarks, with shape (n, p, 2).
            n is the number of samples, p is the number of landmarks,
            and 2 is respectively the y and x coordinates of the landmarks.
        pred_landmarks : torch.Tensor
            The predicted landmarks, with shape (n, p, 2).
            n is the number of samples, p is the number of landmarks,
            and 2 is respectively the y and x coordinates of the landmarks.
        dim : tuple[int, ...] | torch.Tensor
            The dimension of the image, with shape (2,).
            2 is respectively the height and width of the image.
        dim_orig : torch.Tensor
            The original dimension of the image, with shape (2,).
            2 is respectively the height and width of the image.
        pixel_spacing : torch.Tensor
            The pixel spacing of the image, with shape (2,).
            2 is respectively the height and width of the image.
        padding : torch.Tensor
            The padding of the image, with shape (2,).
            2 is respectively the height and width of the image.
        class_names: list
            The names of the landmarks. The default is None.
        radius : list[float], optional
            The radius of the success detection rate (SDR). The default is [2, 2.5, 3, 4].
        digits : int, optional
            The number of digits to round the statistics. The default is 2.
        unit : str, optional
            The unit of distance. The default is 'mm'.
        output_dict : bool, optional
            Whether to output the detection report as a dictionary. The default is False.

    Returns:
        report : dict
            The detection report. Only output when output_dict is True.
    """
    # Calculate the cumulative point-to-point error (CPE)
    pe = point_error(
        true_landmarks, pred_landmarks, dim, dim_orig, pixel_spacing, padding, reduction="none"
    )
    pe = pe.cpu().detach().numpy()
    if class_names is None:
        class_names = [f"Landmark {i}" for i in range(true_landmarks.shape[1])]
    # Calculate the detection report
    report: dict[str, dict[str, float | dict[float, float]]] = {}
    for i, class_name in enumerate(class_names):  # type: ignore
        report[class_name] = {}
        report[class_name]["Mean"] = np.mean(pe[:, i])
        report[class_name]["Median"] = np.median(pe[:, i])
        report[class_name]["Std"] = np.std(pe[:, i])
        report[class_name]["Min"] = np.min(pe[:, i])
        report[class_name]["Max"] = np.max(pe[:, i])
        sdr_class = sdr(
            radius,
            true_landmarks=true_landmarks[:, i],
            pred_landmarks=pred_landmarks[:, i],
            dim=dim,
            dim_orig=dim_orig,
            pixel_spacing=pixel_spacing,
            padding=padding,
        )
        report[class_name]["SDR"] = sdr_class

    # Print the detection report in a nice table
    sdr_names = [f"SDR (PE≤{r}{unit})" for r in radius]
    print("Detection report:")
    print("1# Point-to-point error (PE) statistics:")
    print("=" * (20 + 10 * 5))
    print(f"{'Class':<20}{'Mean':<10}{'Median':<10}{'Std':<10}{'Min':<10}{'Max':<10}")
    print("-" * (20 + 10 * 5))
    for class_name in report:
        print(
            f"{class_name:<20}"
            f"{report[class_name]['Mean']:<10.{digits}f}"
            f"{report[class_name]['Median']:<10.{digits}f}"
            f"{report[class_name]['Std']:<10.{digits}f}"
            f"{report[class_name]['Min']:<10.{digits}f}"
            f"{report[class_name]['Max']:<10.{digits}f}"
        )
    print("=" * (20 + 10 * 5))

    print("\n2# Success detection rate (SDR):")
    print("=" * (20 + 15 * len(radius)))
    print("".join([f"{'Class':<20}"] + [f"{sdr_name:<15}" for sdr_name in sdr_names]))
    print("-" * (20 + 15 * len(radius)))
    for class_name in report:
        print(
            "".join(
                [f"{class_name:<20}"]
                + [f"{report[class_name]['SDR'][r]:<15.{digits}f}" for r in radius]  # type: ignore
            )
        )
    print("=" * (20 + 15 * len(radius)))

    if output_dict:
        return report


def multi_instance_detection_report(
    true_landmarks: torch.Tensor,
    pred_landmarks: torch.Tensor,
    true_positives: torch.Tensor,
    false_positives: torch.Tensor,
    false_negatives: torch.Tensor,
    dim: tuple[int, ...] | torch.Tensor,
    dim_orig: torch.Tensor,
    pixel_spacing: torch.Tensor,
    padding: torch.Tensor,
    class_names: Optional[list[str]] = None,
    radius: list[float] = [2, 2.5, 3, 4],
    digits: int = 2,
    unit: str = "mm",
    output_dict: bool = False,
):
    """Calculate the multi instance detection report.

    Args:
        true_landmarks : torch.Tensor
            The true landmarks, with shape (n, p, i, 2). n is the number of samples, p is the number
            of classes of landmarks, i is the number of instances, and 2 is respectively the y and
            x coordinates of the landmarks.
        pred_landmarks : torch.Tensor
            The predicted landmarks, with shape (n, p, i, 2). n is the number of samples, p is the
            number of classes of landmarks, i is the number of instances, and 2 is respectively the
            y and x coordinates of the landmarks.
        true_positives : torch.Tensor
            The true positives, with shape (n, p, i). n is the number of samples, p is the number of
            classes of landmarks, and i is the number of instances.
        false_positives : torch.Tensor
            The false positives, with shape (n, p, i). n is the number of samples, p is the number
            of classes of landmarks, and i is the number of instances.
        false_negatives : torch.Tensor
            The false negatives, with shape (n, p, i). n is the number of samples, p is the number
            of classes of landmarks, and i is the number of instances.
        dim : tuple[int, ...] | torch.Tensor
            The dimension of the image, with shape (2,).
            2 is respectively the height and width of the image.
        dim_orig : torch.Tensor
            The original dimension of the image, with shape (2,).
            2 is respectively the height and width of the image.
        pixel_spacing : torch.Tensor
            The pixel spacing of the image, with shape (2,).
            2 is respectively the height and width of the image.
        padding : torch.Tensor
            The padding of the image, with shape (2,).
            2 is respectively the height and width of the image.
        class_names: list
            The names of the landmarks. The default is None.
        radius : list[float], optional
            The radius of the success detection rate (SDR). The default is [2, 2.5, 3, 4].
        digits : int, optional
            The number of digits to round the statistics. The default is 2.
        unit : str, optional
            The unit of distance. The default is 'mm'.
        output_dict : bool, optional
            Whether to output the detection report as a dictionary. The default is False.

    Returns:
        report : dict
            The detection report. Only output when output_dict is True.
    """
    # Calculate the cumulative point-to-point error (CPE)
    pe = point_error(
        true_landmarks, pred_landmarks, dim, dim_orig, pixel_spacing, padding, reduction="none"
    )
    pe = pe.cpu().detach().numpy()

    # Calculate the detection report
    report: dict[str, dict[str, float | dict[float, float]]] = {}
    for i, class_name in enumerate(class_names):  # type: ignore
        report[class_name] = {}
        report[class_name]["Mean"] = np.nanmean(pe[:, i])
        report[class_name]["Median"] = np.nanmedian(pe[:, i])
        report[class_name]["Std"] = np.nanstd(pe[:, i])
        report[class_name]["Min"] = np.nanmin(pe[:, i])
        report[class_name]["Max"] = np.nanmax(pe[:, i])
        report[class_name]["TP"] = torch.sum(true_positives[:, i]).item()
        report[class_name]["FP"] = torch.sum(false_positives[:, i]).item()
        report[class_name]["FN"] = torch.sum(false_negatives[:, i]).item()
        report[class_name]["TPR"] = report[class_name]["TP"] / (  # type: ignore
            report[class_name]["TP"] + report[class_name]["FN"]  # type: ignore
        )
        sdr_class = sdr(
            radius,
            true_landmarks=true_landmarks[:, i],
            pred_landmarks=pred_landmarks[:, i],
            dim=dim,
            dim_orig=dim_orig,
            pixel_spacing=pixel_spacing,
            padding=padding,
        )
        report[class_name]["SDR"] = sdr_class

    # Print the detection report in a nice table
    sdr_names = [f"SDR (PE≤{r}{unit})" for r in radius]
    print("Detection report:")
    print("1# Instance detection statistics:")
    print("=" * (20 + 10 * 4))
    print(f"{'Class':<20}{'TP':<10}{'FP':<10}{'FN':<10}{'TPR':<10}")
    print("-" * (20 + 10 * 4))
    for class_name in report:
        print(
            f"{class_name:<20}"
            f"{report[class_name]['TP']:<10}"
            f"{report[class_name]['FP']:<10}"
            f"{report[class_name]['FN']:<10}"
            f"{report[class_name]['TPR']:<10.{digits}f}"
        )
    print("=" * (20 + 10 * 4))

    print("\n2# Point-to-point error (PE) statistics:")
    print("=" * (20 + 10 * 5))
    print(f"{'Class':<20}{'Mean':<10}{'Median':<10}{'Std':<10}{'Min':<10}{'Max':<10}")
    print("-" * (20 + 10 * 5))
    for class_name in report:
        print(
            f"{class_name:<20}"
            f"{report[class_name]['Mean']:<10.{digits}f}"
            f"{report[class_name]['Median']:<10.{digits}f}"
            f"{report[class_name]['Std']:<10.{digits}f}"
            f"{report[class_name]['Min']:<10.{digits}f}"
            f"{report[class_name]['Max']:<10.{digits}f}"
        )
    print("=" * (20 + 10 * 5))

    print("\n3# Success detection rate (SDR):")
    print("=" * (20 + 15 * len(radius)))
    print("".join([f"{'Class':<20}"] + [f"{sdr_name:<15}" for sdr_name in sdr_names]))
    print("-" * (20 + 15 * len(radius)))
    for class_name in report:
        print(
            "".join(
                [f"{class_name:<20}"]
                + [f"{report[class_name]['SDR'][r]:<15.{digits}f}" for r in radius]  # type: ignore
            )
        )
    print("=" * (20 + 15 * len(radius)))

    if output_dict:
        return report



"""
Landmark transforms.
"""

from typing import Optional

import torch


def resize_landmarks(
    landmarks: torch.Tensor,
    orig_dim: tuple[int, ...] | torch.Tensor,
    new_dim: tuple[int, ...] | torch.Tensor,
    padding: Optional[tuple[int, ...] | torch.Tensor] = None,
) -> torch.Tensor:
    """
    Resize landmarks to ``new_dim`` and apply padding if necessary (supplied).

    Args:
        landmarks (torch.Tensor): landmarks to resize.
        orig_dim (tuple[int, int]): original dimension of the images.
        new_dim (tuple[int, int]): new dimension of the images.
        padding (tuple[int, int]): padding applied to the images.

    Returns:
        landmarks (torch.Tensor): resized and padded code landmarks.
    """
    orig_dim = torch.Tensor(orig_dim)
    new_dim = torch.Tensor(new_dim)
    if padding is None:
        padding = torch.zeros(len(orig_dim.size()))
    else:
        padding = torch.Tensor(padding)
    t_landmarks = landmarks + padding
    t_landmarks = t_landmarks * (new_dim / (orig_dim + 2 * padding))
    return t_landmarks


def affine_landmarks(
    landmarks: torch.Tensor,
    affine_matrix: torch.Tensor,  # push affine matrix
) -> torch.Tensor:
    """
    Apply an affine transform to landmarks. The affine matrix is assumed to be a push affine
    transform.

    Args:
        landmarks (torch.Tensor): landmarks to transform.
        affine_matrix (torch.Tensor): 4 by 4 affine matrix to apply to the landmarks.

    Returns:
        landmarks (torch.Tensor): transformed landmarks.
    """
    assert affine_matrix.shape == (4, 4)
    # push affine transform
    if landmarks.shape[-1] == 2:
        landmarks_t = affine_matrix @ torch.cat(
            [
                landmarks,
                torch.zeros((*landmarks.shape[:-1], 1)),
                torch.ones((*landmarks.shape[:-1], 1)),
            ],
            dim=-1,
        ).unsqueeze(-1)
        return landmarks_t[..., :2, 0]
    elif landmarks.shape[-1] == 3:
        landmarks_t = affine_matrix @ torch.cat(
            [
                landmarks,
                torch.ones((*landmarks.shape[:-1], 1)),
            ],
            dim=-1,
        ).unsqueeze(-1)
        return landmarks_t[..., :3, 0]
    raise ValueError(f"Landmarks shape {landmarks.shape} not supported")


def elastic_deform_landmarks(
    landmarks: torch.Tensor,
    grid_displacement: torch.Tensor,
) -> torch.Tensor:
    """
    Apply an elastic deformation to landmarks.

    Args:
        landmarks (torch.Tensor): landmarks to transform.
        displacement (torch.Tensor): displacement field to apply to the landmarks.

    Returns:
        landmarks (torch.Tensor): transformed landmarks.
    """
    raise NotImplementedError(
        "Elastic deformation not implemented yet. MONAI does not"
        " support backtracking the grid sample and affine transform."
    )
    return landmarks



"""
This module contains the functions to load the ISBI 2015 cephalometric landmark detection challenge
dataset.
"""

import glob
import os
import zipfile

import numpy as np
import opendatasets as od  # type: ignore
import pandas as pd  # type: ignore
import rarfile  # type: ignore

from landmarker.data.landmark_dataset import (
    HeatmapDataset,
    LandmarkDataset,
    MaskDataset,
    PatchDataset,
    PatchMaskDataset,
)


def get_cepha_dataset(path_dir: str, junior: bool = False, cv: bool = True):
    """Returns the paths to the images and the landmarks of the CEPH dataset from the ISBI 2014 &
    2015 challenges. But not from the kaggle dataset but from this repository:
        https://figshare.com/s/37ec464af8e81ae6ebbf?file=5466581

    Args:
        path_dir (str): The path to the directory where the dataset should be stored.
        junior (bool, optional): Whether to use the junior or the average of senior and junior annotator. Defaults to False.
        cv (bool, optional): Whether to use the cross validation splits from the paper. Defaults to
            True.
    """
    if not os.path.exists(path_dir + "/ISBI2015"):
        od.download(
            "https://figshare.com/ndownloader/articles/3471833?private_link=37ec464af8e81ae6ebbf",
            path_dir,
        )
        # Change the name of the folder to ISBI2015
        with zipfile.ZipFile(path_dir + "/3471833.zip", "r") as zip_ref:
            zip_ref.extractall(path=path_dir + "/ISBI2015")
        os.remove(path_dir + "/3471833.zip")
        try:
            with rarfile.RarFile(path_dir + "/ISBI2015/AnnotationsByMD.rar", "r") as zip_ref:
                zip_ref.extractall(path=path_dir + "/ISBI2015")
            with rarfile.RarFile(path_dir + "/ISBI2015/RawImage.rar", "r") as zip_ref:
                zip_ref.extractall(path=path_dir + "/ISBI2015")
        except rarfile.BadRarFile:
            raise rarfile.BadRarFile(
                "Possibly you need to install unrar. On linux: sudo apt install unrar"
            )
        os.remove(path_dir + "/ISBI2015/AnnotationsByMD.rar")
        os.remove(path_dir + "/ISBI2015/RawImage.rar")
        os.remove(path_dir + "/ISBI2015/EvaluationCode.rar")
        os.mkdir(path_dir + "/ISBI2015/cv_payer")
        for i in range(1, 5):
            od.download(
                "https://raw.githubusercontent.com/christianpayer/MedicalDataAugmentationTool"
                + f"-HeatmapUncertainty/main/setup_ann/all_landmarks/cv/{i}.txt",
                path_dir + "/ISBI2015/cv_payer",
            )

    landmarks_list = []
    for i in range(400):
        landmarks_list.append(
            pd.read_csv(
                path_dir + f"/ISBI2015/400_junior/{str(i+1).zfill(3)}.txt",
                sep=",",
                header=None,
            )[:19].to_numpy()
        )
    junior_landmarks = np.concatenate(landmarks_list, axis=0).reshape((-1, 19, 2))
    junior_landmarks = np.flip(junior_landmarks, axis=-1)
    if junior:
        landmarks = junior_landmarks
    else:
        landmarks_list = []
        for i in range(400):
            landmarks_list.append(
                pd.read_csv(
                    path_dir + f"/ISBI2015/400_senior/{str(i+1).zfill(3)}.txt",
                    sep=",",
                    header=None,
                )[:19].to_numpy()
            )
        senior_landmarks = np.concatenate(landmarks_list, axis=0).reshape((-1, 19, 2))
        senior_landmarks = np.flip(senior_landmarks, axis=-1)
        landmarks = (junior_landmarks + senior_landmarks) / 2

    if cv:
        indices_fold_1 = (
            pd.read_table(path_dir + f"/ISBI2015/cv_payer/{1}.txt", header=None)
            .to_numpy()
            .flatten()
            .tolist()
        )
        indices_fold_2 = (
            pd.read_table(path_dir + f"/ISBI2015/cv_payer/{2}.txt", header=None)
            .to_numpy()
            .flatten()
            .tolist()
        )
        indices_fold_3 = (
            pd.read_table(path_dir + f"/ISBI2015/cv_payer/{3}.txt", header=None)
            .to_numpy()
            .flatten()
            .tolist()
        )
        indices_fold_4 = (
            pd.read_table(path_dir + f"/ISBI2015/cv_payer/{4}.txt", header=None)
            .to_numpy()
            .flatten()
            .tolist()
        )
        image_paths_fold1 = [
            glob.glob(path_dir + f"/ISBI2015/RawImage/*/{str(i).zfill(3)}.bmp")[0]
            for i in indices_fold_1
        ]
        image_paths_fold2 = [
            glob.glob(path_dir + f"/ISBI2015/RawImage/*/{str(i).zfill(3)}.bmp")[0]
            for i in indices_fold_2
        ]
        image_paths_fold3 = [
            glob.glob(path_dir + f"/ISBI2015/RawImage/*/{str(i).zfill(3)}.bmp")[0]
            for i in indices_fold_3
        ]
        image_paths_fold4 = [
            glob.glob(path_dir + f"/ISBI2015/RawImage/*/{str(i).zfill(3)}.bmp")[0]
            for i in indices_fold_4
        ]
        return (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks[[i - 1 for i in indices_fold_1]],
            landmarks[[i - 1 for i in indices_fold_2]],
            landmarks[[i - 1 for i in indices_fold_3]],
            landmarks[[i - 1 for i in indices_fold_4]],
            np.array([[0.1, 0.1]]).repeat(len(image_paths_fold1), axis=0),
            np.array([[0.1, 0.1]]).repeat(len(image_paths_fold2), axis=0),
            np.array([[0.1, 0.1]]).repeat(len(image_paths_fold3), axis=0),
            np.array([[0.1, 0.1]]).repeat(len(image_paths_fold4), axis=0),
        )

    image_paths_train = [
        path_dir + f"/ISBI2015/RawImage/TrainingData/{str(i).zfill(3)}.bmp" for i in range(1, 151)
    ]
    landmarks_train = landmarks[:150]
    pixel_spacings_train = np.array([[0.1, 0.1]]).repeat(len(landmarks_train), axis=0)
    image_paths_test1 = [
        path_dir + f"/ISBI2015/RawImage/Test1Data/{str(i).zfill(3)}.bmp" for i in range(151, 301)
    ]
    landmarks_test1 = landmarks[150:300]
    pixel_spacings_test1 = np.array([[0.1, 0.1]]).repeat(len(landmarks_test1), axis=0)
    image_paths_test2 = [
        path_dir + f"/ISBI2015/RawImage/Test2Data/{str(i).zfill(3)}.bmp" for i in range(301, 401)
    ]
    landmarks_test2 = landmarks[300:]
    pixel_spacings_test2 = np.array([[0.1, 0.1]]).repeat(len(landmarks_test2), axis=0)
    return (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    )


def get_cepha_landmark_datasets(
    path_dir: str,
    train_transform=None,
    inference_transform=None,
    store_imgs=True,
    dim_img=None,
    junior=False,
    single_dataset=False,
    cv=False,
) -> (
    LandmarkDataset
    | tuple[LandmarkDataset, LandmarkDataset, LandmarkDataset]
    | tuple[LandmarkDataset, LandmarkDataset, LandmarkDataset, LandmarkDataset]
):
    """Returns a LandmarkDataset objects with the CEPH dataset, a combination of the ISBI 2014 &
    2015 challenges. The dataset is split into train, test1 and test2. The same approach as in
    "CephaNN: A Multi-Head Attention Network for Cephalometric Landmark Detection" - JIAHOONG QIAN
        et al. is used.
    """
    if single_dataset and cv:
        raise ValueError("Cannot have single dataset and cross validation at the same time.")
    if cv:
        (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks_fold1,
            landmarks_fold2,
            landmarks_fold3,
            landmarks_fold4,
            pixel_spacings_fold1,
            pixel_spacings_fold2,
            pixel_spacings_fold3,
            pixel_spacings_fold4,
        ) = get_cepha_dataset(path_dir, junior=junior, cv=True)
        return (
            LandmarkDataset(
                image_paths_fold1,
                landmarks_fold1,
                pixel_spacing=pixel_spacings_fold1,
                transform=train_transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
            ),
            LandmarkDataset(
                image_paths_fold2,
                landmarks_fold2,
                pixel_spacing=pixel_spacings_fold2,
                transform=train_transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
            ),
            LandmarkDataset(
                image_paths_fold3,
                landmarks_fold3,
                pixel_spacing=pixel_spacings_fold3,
                transform=train_transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
            ),
            LandmarkDataset(
                image_paths_fold4,
                landmarks_fold4,
                pixel_spacing=pixel_spacings_fold4,
                transform=train_transform,
                store_imgs=store_imgs,
                dim_img=dim_img,
            ),
        )
    (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    ) = get_cepha_dataset(path_dir, junior=junior, cv=False)
    if single_dataset:
        return LandmarkDataset(
            image_paths_train + image_paths_test1 + image_paths_test2,
            np.concatenate([landmarks_train, landmarks_test1, landmarks_test2], axis=0),
            pixel_spacing=np.concatenate(
                [pixel_spacings_train, pixel_spacings_test1, pixel_spacings_test2], axis=0
            ),
            transform=train_transform,
            store_imgs=store_imgs,
            dim_img=dim_img,
        )
    return (
        LandmarkDataset(
            image_paths_train,
            landmarks_train,
            pixel_spacing=pixel_spacings_train,
            transform=train_transform,
            store_imgs=store_imgs,
            dim_img=dim_img,
        ),
        LandmarkDataset(
            image_paths_test1,
            landmarks_test1,
            pixel_spacing=pixel_spacings_test1,
            transform=inference_transform,
            store_imgs=store_imgs,
            dim_img=dim_img,
        ),
        LandmarkDataset(
            image_paths_test2,
            landmarks_test2,
            pixel_spacing=pixel_spacings_test2,
            transform=inference_transform,
            store_imgs=store_imgs,
            dim_img=dim_img,
        ),
    )


def get_cepha_heatmap_datasets(
    path_dir: str,
    train_transform=None,
    inference_transform=None,
    sigma: float = 1,
    junior: bool = False,
    single_dataset: bool = False,
    cv: bool = False,
    **kwargs,
) -> (
    HeatmapDataset
    | tuple[HeatmapDataset, HeatmapDataset, HeatmapDataset]
    | tuple[HeatmapDataset, HeatmapDataset, HeatmapDataset, HeatmapDataset]
):
    """Returns a HeatmapDataset with the ISBI 2015 cephalogram challenge dataset. The dataset is
    split into train, test1 and test2. The same approach as in "CephaNN: A Multi-Head Attention
    Network for Cephalometric Landmark Detection" - JIAHOONG QIAN et al. is used.

    Args:
        path_dir (str): The path to the directory where the dataset should be stored.
        train_transform (Optional[Callable], optional): A transformation to apply to the images and
            landmarks during training. Defaults to None.
        inference_transform (Optional[Callable], optional): A transformation to apply to the images
            and landmarks during inference. Defaults to None.
        sigma (int, optional): The sigma value for the gaussian kernel. Defaults to 1.
        junior (bool, optional): Whether to use the junior or senior annotator. Defaults to False.
        single_dataset (bool, optional): Whether to return a single dataset with all images and
            landmarks. Defaults to False.
        **kwargs: Additional keyword arguments for the HeatmapDataset.
    """
    if single_dataset and cv:
        raise ValueError("Cannot have single dataset and cross validation at the same time.")
    if cv:
        (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks_fold1,
            landmarks_fold2,
            landmarks_fold3,
            landmarks_fold4,
            pixel_spacings_fold1,
            pixel_spacings_fold2,
            pixel_spacings_fold3,
            pixel_spacings_fold4,
        ) = get_cepha_dataset(path_dir, junior=junior, cv=True)
        return (
            HeatmapDataset(
                image_paths_fold1,
                landmarks_fold1,
                pixel_spacing=pixel_spacings_fold1,
                transform=train_transform,
                sigma=sigma,
                **kwargs,
            ),
            HeatmapDataset(
                image_paths_fold2,
                landmarks_fold2,
                pixel_spacing=pixel_spacings_fold2,
                transform=train_transform,
                sigma=sigma,
                **kwargs,
            ),
            HeatmapDataset(
                image_paths_fold3,
                landmarks_fold3,
                pixel_spacing=pixel_spacings_fold3,
                transform=train_transform,
                sigma=sigma,
                **kwargs,
            ),
            HeatmapDataset(
                image_paths_fold4,
                landmarks_fold4,
                pixel_spacing=pixel_spacings_fold4,
                transform=train_transform,
                sigma=sigma,
                **kwargs,
            ),
        )
    (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    ) = get_cepha_dataset(path_dir, junior=junior, cv=False)
    if single_dataset:
        return HeatmapDataset(
            image_paths_train + image_paths_test1 + image_paths_test2,
            np.concatenate([landmarks_train, landmarks_test1, landmarks_test2], axis=0),
            pixel_spacing=np.concatenate(
                [pixel_spacings_train, pixel_spacings_test1, pixel_spacings_test2], axis=0
            ),
            transform=train_transform,
            sigma=sigma,
            **kwargs,
        )
    return (
        HeatmapDataset(
            image_paths_train,
            landmarks_train,
            pixel_spacing=pixel_spacings_train,
            transform=train_transform,
            sigma=sigma,
            **kwargs,
        ),
        HeatmapDataset(
            image_paths_test1,
            landmarks_test1,
            pixel_spacing=pixel_spacings_test1,
            transform=inference_transform,
            sigma=sigma,
            **kwargs,
        ),
        HeatmapDataset(
            image_paths_test2,
            landmarks_test2,
            pixel_spacing=pixel_spacings_test2,
            transform=inference_transform,
            sigma=sigma,
            **kwargs,
        ),
    )


def get_cepha_mask_datasets(
    path_dir: str,
    train_transform=None,
    inference_transform=None,
    junior: bool = False,
    single_dataset: bool = False,
    cv: bool = False,
    **kwargs,
) -> (
    MaskDataset
    | tuple[MaskDataset, MaskDataset, MaskDataset]
    | tuple[MaskDataset, MaskDataset, MaskDataset, MaskDataset]
):
    """Returns a MaskDataset with the ISBI 2015 cephalogram challenge dataset. The dataset is
    split into train, test1 and test2. The same approach as in "CephaNN: A Multi-Head Attention
    Network for Cephalometric Landmark Detection" - JIAHOONG QIAN et al. is used.

    Args:
        path_dir (str): The path to the directory where the dataset should be stored.
        train_transform (Optional[Callable], optional): A transformation to apply to the images and
            masks during training. Defaults to None.
        inference_transform (Optional[Callable], optional): A transformation to apply to the images
            and masks during inference. Defaults to None.
        junior (bool, optional): Whether to use the junior or senior annotator. Defaults to False.
        single_dataset (bool, optional): Whether to return a single dataset with all images and
            landmarks. Defaults to False.
        **kwargs: Additional keyword arguments for the MaskDataset.
    """
    if single_dataset and cv:
        raise ValueError("Cannot have single dataset and cross validation at the same time.")
    if cv:
        (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks_fold1,
            landmarks_fold2,
            landmarks_fold3,
            landmarks_fold4,
            pixel_spacings_fold1,
            pixel_spacings_fold2,
            pixel_spacings_fold3,
            pixel_spacings_fold4,
        ) = get_cepha_dataset(path_dir, junior=junior, cv=True)
        return (
            MaskDataset(
                image_paths_fold1,
                landmarks_fold1,
                pixel_spacing=pixel_spacings_fold1,
                transform=train_transform,
                **kwargs,
            ),
            MaskDataset(
                image_paths_fold2,
                landmarks_fold2,
                pixel_spacing=pixel_spacings_fold2,
                transform=train_transform,
                **kwargs,
            ),
            MaskDataset(
                image_paths_fold3,
                landmarks_fold3,
                pixel_spacing=pixel_spacings_fold3,
                transform=train_transform,
                **kwargs,
            ),
            MaskDataset(
                image_paths_fold4,
                landmarks_fold4,
                pixel_spacing=pixel_spacings_fold4,
                transform=train_transform,
                **kwargs,
            ),
        )
    (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    ) = get_cepha_dataset(path_dir, junior=junior, cv=False)
    if single_dataset:
        return MaskDataset(
            image_paths_train + image_paths_test1 + image_paths_test2,
            np.concatenate([landmarks_train, landmarks_test1, landmarks_test2], axis=0),
            pixel_spacing=np.concatenate(
                [pixel_spacings_train, pixel_spacings_test1, pixel_spacings_test2], axis=0
            ),
            transform=train_transform,
            **kwargs,
        )
    return (
        MaskDataset(
            image_paths_train,
            landmarks_train,
            pixel_spacing=pixel_spacings_train,
            transform=train_transform,
            **kwargs,
        ),
        MaskDataset(
            image_paths_test1,
            landmarks_test1,
            pixel_spacing=pixel_spacings_test1,
            transform=inference_transform,
            **kwargs,
        ),
        MaskDataset(
            image_paths_test2,
            landmarks_test2,
            pixel_spacing=pixel_spacings_test2,
            transform=inference_transform,
            **kwargs,
        ),
    )


def get_cepha_patch_datasets(
    path_dir: str,
    index_landmark: int = 0,
    train_transform=None,
    inference_transform=None,
    store_imgs=True,
    junior=False,
    single_dataset=False,
    cv=False,
    **kwargs,
):
    """Returns a PatchDataset objects with the CEPH dataset, a combination of the ISBI 2014 &
    2015 challenges. The dataset is split into train, test1 and test2. The same approach as in
    "CephaNN: A Multi-Head Attention Network for Cephalometric Landmark Detection" - JIAHOONG QIAN
        et al. is used.
    """
    if single_dataset and cv:
        raise ValueError("Cannot have single dataset and cross validation at the same time.")
    if cv:
        (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks_fold1,
            landmarks_fold2,
            landmarks_fold3,
            landmarks_fold4,
            pixel_spacings_fold1,
            pixel_spacings_fold2,
            pixel_spacings_fold3,
            pixel_spacings_fold4,
        ) = get_cepha_dataset(path_dir, junior=junior, cv=True)
        return (
            PatchDataset(
                image_paths_fold1,
                landmarks_fold1,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold1,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchDataset(
                image_paths_fold2,
                landmarks_fold2,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold2,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchDataset(
                image_paths_fold3,
                landmarks_fold3,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold3,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchDataset(
                image_paths_fold4,
                landmarks_fold4,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold4,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
        )
    (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    ) = get_cepha_dataset(path_dir, junior=junior, cv=False)
    if single_dataset:
        return PatchDataset(
            image_paths_train + image_paths_test1 + image_paths_test2,
            np.concatenate([landmarks_train, landmarks_test1, landmarks_test2], axis=0),
            index_landmark=index_landmark,
            pixel_spacing=np.concatenate(
                [pixel_spacings_train, pixel_spacings_test1, pixel_spacings_test2], axis=0
            ),
            transform=train_transform,
            store_imgs=store_imgs,
            **kwargs,
        )
    return (
        PatchDataset(
            image_paths_train,
            landmarks_train,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_train,
            transform=train_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
        PatchDataset(
            image_paths_test1,
            landmarks_test1,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_test1,
            transform=inference_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
        PatchDataset(
            image_paths_test2,
            landmarks_test2,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_test2,
            transform=inference_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
    )


def get_cepha_patch_mask_datasets(
    path_dir: str,
    index_landmark: int = 0,
    train_transform=None,
    inference_transform=None,
    store_imgs=True,
    junior=False,
    single_dataset=False,
    cv=False,
    **kwargs,
):
    """Returns a PatchMaskDataset objects with the CEPH dataset, a combination of the ISBI 2014 &
    2015 challenges. The dataset is split into train, test1 and test2. The same approach as in
    "CephaNN: A Multi-Head Attention Network for Cephalometric Landmark Detection" - JIAHOONG QIAN
        et al. is used.
    """
    if single_dataset and cv:
        raise ValueError("Cannot have single dataset and cross validation at the same time.")
    if cv:
        (
            image_paths_fold1,
            image_paths_fold2,
            image_paths_fold3,
            image_paths_fold4,
            landmarks_fold1,
            landmarks_fold2,
            landmarks_fold3,
            landmarks_fold4,
            pixel_spacings_fold1,
            pixel_spacings_fold2,
            pixel_spacings_fold3,
            pixel_spacings_fold4,
        ) = get_cepha_dataset(path_dir, junior=junior, cv=True)
        return (
            PatchMaskDataset(
                image_paths_fold1,
                landmarks_fold1,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold1,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchMaskDataset(
                image_paths_fold2,
                landmarks_fold2,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold2,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchMaskDataset(
                image_paths_fold3,
                landmarks_fold3,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold3,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
            PatchMaskDataset(
                image_paths_fold4,
                landmarks_fold4,
                index_landmark=index_landmark,
                pixel_spacing=pixel_spacings_fold4,
                transform=train_transform,
                store_imgs=store_imgs,
                **kwargs,
            ),
        )
    (
        image_paths_train,
        image_paths_test1,
        image_paths_test2,
        landmarks_train,
        landmarks_test1,
        landmarks_test2,
        pixel_spacings_train,
        pixel_spacings_test1,
        pixel_spacings_test2,
    ) = get_cepha_dataset(path_dir, junior=junior, cv=False)
    if single_dataset:
        return PatchMaskDataset(
            image_paths_train + image_paths_test1 + image_paths_test2,
            np.concatenate([landmarks_train, landmarks_test1, landmarks_test2], axis=0),
            index_landmark=index_landmark,
            pixel_spacing=np.concatenate(
                [pixel_spacings_train, pixel_spacings_test1, pixel_spacings_test2], axis=0
            ),
            transform=train_transform,
            store_imgs=store_imgs,
            **kwargs,
        )
    return (
        PatchMaskDataset(
            image_paths_train,
            landmarks_train,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_train,
            transform=train_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
        PatchMaskDataset(
            image_paths_test1,
            landmarks_test1,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_test1,
            transform=inference_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
        PatchMaskDataset(
            image_paths_test2,
            landmarks_test2,
            index_landmark=index_landmark,
            pixel_spacing=pixel_spacings_test2,
            transform=inference_transform,
            store_imgs=store_imgs,
            **kwargs,
        ),
    )



"""
Vizualization module.
"""

from .evaluation import detection_report, plot_cpe
from .utils import inspection_plot, prediction_inspect_plot

__all__ = ["plot_cpe", "inspection_plot", "detection_report", "prediction_inspect_plot"]



from landmarker.uncertainty import (  # type: ignore
    MR2C2R,
    MR2CCP,
    ConformalRegressorBonferroni,
    ConformalRegressorMahalanobis,
    ConformalRegressorMaxNonconformity,
    ContourHuggingRegressor,
    MultivariateNormalRegressor,
    resize_landmarks,
    transform_heatmap_to_original_size,
    transform_heatmap_to_original_size_numpy,
)


def test_import_uncertainty():
    """Test that the uncertainty module can be imported correctly."""
    assert MR2C2R is not None
    assert MR2CCP is not None
    assert ConformalRegressorBonferroni is not None
    assert ConformalRegressorMahalanobis is not None
    assert ConformalRegressorMaxNonconformity is not None
    assert ContourHuggingRegressor is not None
    assert MultivariateNormalRegressor is not None
    assert resize_landmarks is not None
    assert transform_heatmap_to_original_size is not None
    assert transform_heatmap_to_original_size_numpy is not None



"""Spatial Configuration Network (SCN) from the paper
"Integrating spatial configuration into heatmap regression based CNNs for landmark localization"
by Payer et al. (2019). https://www.sciencedirect.com/science/article/pii/S1361841518305784"""

from typing import Sequence

import torch
from monai.networks.nets import UNet
from torch import nn


class SpatialConfigurationNet(nn.Module):
    """
    Adapted implementation of the Spatial Configuration Network (SCN) from the paper
    "Integrating spatial configuration into heatmap regression based CNNs for landmark localization"
    by Payer et al. (2019).
    https://www.sciencedirect.com/science/article/pii/S1361841518305784

    Args:
        spatial_dims (int): number of spatial dimensions of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        la_channels (Sequence[int], optional): number of output channels for each convolutional
            layer.
        la_kernel_size (int, optional): kernel size for the convolutional layers.
        la_strides (Sequence[int], optional): strides for the convolutional layers.
        la_num_res_units (int, optional): number of residual units in the convolutional layers.
        la_norm (str, optional): type of normalization to use. Defaults to "INSTANCE".
        la_activation (str, optional): type of activation to use. Defaults to "PRELU".
        la_adn_ordering (str, optional): ordering of the layers in the residual units. Defaults to
            "ADN".
        la_dropout (float, optional): dropout probability. Defaults to 0.0.
        sp_channels (int, optional): number of channels for the convolutional layers.
        sp_kernel_size (int, optional): kernel size for the convolutional layers.
        sp_downsample (int, optional): factor by which the image is downsampled.
        sp_image_input (bool, optional): whether to use the input image as input for the spatial
            configuration network.
    """

    def __init__(
        self,
        spatial_dims: int = 2,
        in_channels: int = 1,
        out_channels: int = 4,
        la_channels: Sequence[int] = (128, 128, 128, 128),
        la_kernel_size: int | tuple[int, int] = 3,
        # We use strides instead of max_pooling as in the paper as it is superior.
        # https://stats.stackexchange.com/questions/387482/pooling-vs-stride-for-downsampling
        la_strides: Sequence[int] = (2, 2, 2),
        la_num_res_units: int = 2,
        la_norm: str = "INSTANCE",
        la_activation: str = "PRELU",
        la_adn_ordering: str = "ADN",
        la_dropout: float = 0.0,  # In the paper, they use 0.5 dropout, however,
        # this seems to be too much, since we use residual connections.
        sp_channels: int = 128,
        sp_kernel_size: int = 11,
        sp_downsample: int = 16,
        sp_image_input: bool = True,
    ):
        super().__init__()
        self.sp_image_input = sp_image_input
        self.la_net = UNet(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=out_channels,
            channels=la_channels,
            strides=la_strides,
            num_res_units=la_num_res_units,
            norm=la_norm,
            dropout=la_dropout,
            act=la_activation,
            kernel_size=la_kernel_size,
            adn_ordering=la_adn_ordering,
        )
        self.sc_net = nn.Sequential(
            nn.AvgPool2d(kernel_size=sp_downsample),
            nn.Conv2d(
                in_channels=out_channels + int(self.sp_image_input) * in_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=out_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.Upsample(scale_factor=sp_downsample, mode="bicubic", align_corners=False),
            nn.Tanh(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)

        Returns:
            torch.Tensor: output tensor of shape (batch_size, out_channels, *img_dims)
        """
        out_la = self.la_net(x)
        if self.sp_image_input:
            out_sc = self.sc_net(torch.cat((x, out_la), dim=1))
        else:
            out_sc = self.sc_net(out_la)
        out = out_la * out_sc
        return out


class OriginalSpatialConfigurationNet(nn.Module):
    """
    Implementation of the Spatial Configuration Network (SCN) from the paper
    "Integrating spatial configuration into heatmap regression based CNNs for landmark localization"
    by Payer et al. (2019).
    https://www.sciencedirect.com/science/article/pii/S1361841518305784

    Args:
        in_channels (int, optional): number of input channels. Defaults to 1.
        out_channels (int, optional): number of output channels. Defaults to 4.
        la_channels (int, optional): number of output channels for each convolutional layer.
            Defaults to 128.
        la_depth (int, optional): number of convolutional layers. Defaults to 3.
        la_kernel_size (int, optional): kernel size for the convolutional layers. Defaults to 3.
        la_dropout (float, optional): dropout probability. Defaults to 0.5.
        sp_channels (int, optional): number of channels for the convolutional layers. Defaults to
            128.
        sp_kernel_size (int, optional): kernel size for the convolutional layers. Defaults to 11.
        sp_downsample (int, optional): factor by which the image is downsampled. Defaults to 16.
        init_weights (bool, optional): whether to initialize the weights of the convolutional
            layers.
    """

    def __init__(
        self,
        in_channels: int = 1,
        out_channels: int = 4,
        la_channels: int = 128,
        la_depth: int = 3,
        la_kernel_size: int | tuple[int, ...] = 3,
        la_dropout: float = 0.5,
        sp_channels: int = 128,
        sp_kernel_size: int = 11,
        sp_downsample: int = 16,
        init_weigths: bool = False,
        spatial_dim: int = 2,
    ) -> None:
        super().__init__()
        self.init_weights = init_weigths
        if spatial_dim == 2:
            conv = nn.Conv2d  # type: ignore
            avg_pool = nn.AvgPool2d  # type: ignore
            mode = "bilinear"
        elif spatial_dim == 3:
            conv = nn.Conv3d  # type: ignore
            avg_pool = nn.AvgPool3d  # type: ignore
            mode = "trilinear"
        else:
            raise ValueError(f"spatial_dim must be 2 or 3, got {spatial_dim}.")
        self.first_conv = nn.Sequential(
            conv(
                in_channels=in_channels,
                out_channels=la_channels,
                kernel_size=la_kernel_size,
                stride=1,
                padding="same",
                bias=True,
            ),
            nn.LeakyReLU(negative_slope=0.1),
        )

        self.la_downlayers = nn.ModuleList(
            [
                DownLayer(
                    in_channels=la_channels,
                    out_channels=la_channels,
                    dropout=la_dropout,
                    kernel_size=la_kernel_size,
                    spatial_dim=spatial_dim,
                )
                for _ in range(la_depth)
            ]
        )
        self.up_layers = nn.ModuleList(
            [
                UpLayer(
                    in_channels=la_channels,
                    out_channels=la_channels,
                    kernel_size=la_kernel_size,
                    spatial_dim=spatial_dim,
                )
                for _ in range(la_depth)
            ]
        )
        self.bottleneck_layer = nn.Sequential(
            conv(
                in_channels=la_channels,
                out_channels=la_channels,
                kernel_size=la_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            conv(
                in_channels=la_channels,
                out_channels=la_channels,
                kernel_size=la_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
        )
        self.last_layer = conv(
            in_channels=la_channels, out_channels=out_channels, kernel_size=1, padding="same"
        )

        self.sc_net = nn.Sequential(
            avg_pool(kernel_size=sp_downsample),
            conv(
                in_channels=out_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            conv(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            conv(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.LeakyReLU(negative_slope=0.1),
            conv(
                in_channels=sp_channels,
                out_channels=out_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.Upsample(scale_factor=sp_downsample, mode=mode, align_corners=False),
            nn.Tanh(),
        )
        if self.init_weights:
            self.apply(lambda m: self._init_weights(m, mean=0.0, std=0.0001))

    def _init_weights(self, module: nn.Module, mean: float = 0.0, std: float = 0.001) -> None:
        if isinstance(module, (nn.Conv2d, nn.ConvTranspose2d)):
            nn.init.trunc_normal_(module.weight, mean, std)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)

        Returns:
            torch.Tensor: output tensor of shape (batch_size, out_channels, *img_dims)
        """
        out_la = self.first_conv(x)
        skips = []
        for down_layer in self.la_downlayers:
            out_la, skip = down_layer(out_la)
            skips.append(skip)
        out_la = self.bottleneck_layer(out_la)
        for up_layer, skip in zip(self.up_layers, reversed(skips)):
            out_la = up_layer(out_la, skip)
        out_la = self.last_layer(out_la)
        out_sc = self.sc_net(out_la)
        out = out_la * out_sc
        return out


class DownLayer(nn.Module):
    """
    Down layer of the Localisation Network (LA) (UNet) from the original SCN paper.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        dropout: float,
        kernel_size: int | tuple[int, int] = 3,
        bias: bool = True,
        spatial_dim: int = 2,
    ) -> None:
        super().__init__()
        if spatial_dim == 2:
            conv = nn.Conv2d
            avg_pool = nn.AvgPool2d
            dropout_func = nn.Dropout2d
        elif spatial_dim == 3:
            conv = nn.Conv3d
            avg_pool = nn.AvgPool3d
            dropout_func = nn.Dropout3d
        else:
            raise ValueError(f"spatial_dim must be 2 or 3, got {spatial_dim}.")
        self.conv1 = nn.Sequential(
            conv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=1,
                padding="same",
                bias=bias,
            ),
            nn.LeakyReLU(negative_slope=0.1),
        )
        self.dropout = dropout_func(p=dropout)
        self.conv2 = nn.Sequential(
            conv(
                in_channels=out_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=1,
                padding="same",
                bias=bias,
            ),
            nn.LeakyReLU(negative_slope=0.1),
        )
        self.pool = avg_pool(kernel_size=2, stride=2)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        r"""
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)

        Returns:
            tuple[torch.Tensor, torch.Tensor]: output tensor of shape (batch_size, out_channels,
                *img_dims) and skip connection tensor of shape (batch_size, out_channels, *img_dims)
        """
        out = self.conv1(x)
        out = self.dropout(out)
        out = self.conv2(out)
        return self.pool(out), out


class UpLayer(nn.Module):
    """
    Up layer of the Localisation Network (LA) (UNet) from the original SCN paper.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, int] = 3,
        bias: bool = True,
        spatial_dim: int = 2,
    ):
        super().__init__()
        if spatial_dim == 2:
            conv = nn.Conv2d
            mode = "bilinear"
        elif spatial_dim == 3:
            conv = nn.Conv3d
            mode = "trilinear"
        else:
            raise ValueError(f"spatial_dim must be 2 or 3, got {spatial_dim}.")
        self.conv1 = nn.Sequential(
            conv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=1,
                padding="same",
                bias=bias,
            ),
            nn.LeakyReLU(negative_slope=0.1),
        )
        self.upsample = nn.Upsample(scale_factor=2, mode=mode, align_corners=False)

    def forward(self, x: torch.Tensor, skip: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)
            skip: skip connection tensor of shape (batch_size, out_channels, *img_dims)

        Returns:
            torch.Tensor: output tensor of shape (batch_size, out_channels, *img_dims)
        """
        return self.upsample(x) + self.conv1(skip)


class OriginalSpatialConfigurationNet3d(OriginalSpatialConfigurationNet):
    """
    Implementation of the Spatial Configuration Network (SCN) from the paper
    "Integrating spatial configuration into heatmap regression based CNNs for landmark localization"
    by Payer et al. (2019).
    https://www.sciencedirect.com/science/article/pii/S1361841518305784

    This is the 3D version of the original SCN.
    """

    def __init__(
        self,
        in_channels: int = 1,
        out_channels: int = 4,
        la_channels: int = 64,
        la_depth: int = 3,
        la_kernel_size: int | tuple[int, ...] = 3,
        la_dropout: float = 0.5,
        sp_channels: int = 64,
        sp_kernel_size: int = 7,
        sp_downsample: int = 4,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            la_channels=la_channels,
            la_depth=la_depth,
            la_kernel_size=la_kernel_size,
            la_dropout=la_dropout,
            sp_channels=sp_channels,
            sp_kernel_size=sp_kernel_size,
            sp_downsample=sp_downsample,
            spatial_dim=3,
        )


class ProbSpatialConfigurationNet(nn.Module):
    """
    Probabilistic Spatial Configuration Network (PSCN)

    Adapted implementation of the Probabilistic Spatial Configuration Network (PSCN) from the paper
    "Integrating spatial configuration into heatmap regression based CNNs for landmark localization"
    by Payer et al. (2019). This is the same as the Spatial Configuration Network (SCN), but with
    a different last layer. Instead of multiplying the output of the SCN with the output of the
    spatial configuration network, we add them together, since the output of the spatial
    configuration network is a probability distribution in the logit space.

    Args:
        spatial_dims (int, optional): number of spatial dimensions of the input image. Defaults to
            2.
        in_channels (int, optional): number of input channels. Defaults to 1.
        out_channels (int, optional): number of output channels. Defaults to 4.
        la_channels (Sequence[int], optional): number of output channels for each convolutional
            layer. Defaults to (128, 128, 128, 128).
        la_kernel_size (int | tuple[int, int], optional): kernel size for the convolutional
            layers. Defaults to 3.
        la_strides (Sequence[int], optional): strides for the convolutional layers. Defaults to
            (2, 2, 2).
        la_num_res_units (int, optional): number of residual units in the convolutional layers.
            Defaults to 2.
        la_norm (str, optional): type of normalization to use. Defaults to "instance".
        la_activation (str, optional): type of activation to use. Defaults to "PRELU".
        la_adn_ordering (str, optional): ordering of the layers in the residual units. Defaults to
            "NDA".
        la_dropout (float, optional): dropout probability. Defaults to 0.0.
        sp_channels (int, optional): number of channels for the convolutional layers. Defaults to
            128.
        sp_kernel_size (int, optional): kernel size for the convolutional layers. Defaults to 11.
        sp_downsample (int, optional): factor by which the image is downsampled. Defaults to 16.
        sp_image_input (bool, optional): whether to use the input image as input for the spatial
    """

    def __init__(
        self,
        spatial_dims: int = 2,
        in_channels: int = 1,
        out_channels: int = 4,
        la_channels: Sequence[int] = (128, 128, 128, 128, 128),
        la_kernel_size: int | tuple[int, int] = 3,
        # We use strides instead of max_pooling as in the paper as it is superior.
        # https://stats.stackexchange.com/questions/387482/pooling-vs-stride-for-downsampling
        la_strides: Sequence[int] = (2, 2, 2, 2),
        la_num_res_units: int = 2,
        la_norm: str = "instance",
        la_activation: str = "PRELU",
        la_adn_ordering: str = "NDA",
        la_dropout: float = 0.0,  # In the paper, they use 0.5 dropout, however,
        # this seems to be too much, since we use residual connections.
        sp_channels: int = 128,
        sp_kernel_size: int = 11,
        sp_downsample: int = 16,
        sp_image_input: int = True,
    ) -> None:
        super().__init__()
        self.sp_image_input = sp_image_input
        self.la_net = UNet(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=out_channels,
            channels=la_channels,
            strides=la_strides,
            num_res_units=la_num_res_units,
            norm=la_norm,
            dropout=la_dropout,
            act=la_activation,
            kernel_size=la_kernel_size,
            adn_ordering=la_adn_ordering,
        )
        self.sc_net = nn.Sequential(
            nn.AvgPool2d(kernel_size=sp_downsample),
            nn.Conv2d(
                in_channels=out_channels + int(self.sp_image_input) * in_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.PReLU(),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.PReLU(),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=sp_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.PReLU(),
            nn.Conv2d(
                in_channels=sp_channels,
                out_channels=out_channels,
                kernel_size=sp_kernel_size,
                padding="same",
            ),
            nn.Upsample(scale_factor=sp_downsample, mode="bicubic", align_corners=False),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *img_dims)
        Returns:
            torch.Tensor: output tensor of shape (batch_size, out_channels, *img_dims)
        """
        out_la = self.la_net(x)
        if self.sp_image_input:
            out_sc = self.sc_net(torch.cat((x, out_la), dim=1))
        else:
            out_sc = self.sc_net(out_la)
        out = out_la + out_sc
        return out



"""Test landmark transforms."""

import pytest
import torch
from src.landmarker.transforms.landmarks import affine_landmarks, resize_landmarks


def test_resize_landmarks_without_padding():
    # Create a tensor of landmarks
    landmarks = torch.tensor([[1, 2], [3, 4], [5, 6]])

    # Original and new dimensions
    orig_dim = (6, 6)
    new_dim = (12, 12)

    # Call the function
    resized_landmarks = resize_landmarks(landmarks, orig_dim, new_dim)

    # Check that the output has the expected values
    expected_output = torch.tensor([[2, 4], [6, 8], [10, 12]])
    assert torch.all(resized_landmarks.eq(expected_output))

    # Create a multi-instance tensor of landmarks
    landmarks = torch.tensor([[[1, 2], [3, 4], [5, 6]],
                              [[7, 8], [9, 10], [11, 12]]])

    # Call the function
    resized_landmarks = resize_landmarks(landmarks, orig_dim, new_dim)

    # Check that the output has the expected values
    expected_output = torch.tensor([[[2, 4], [6, 8], [10, 12]],
                                    [[14, 16], [18, 20], [22, 24]]])
    assert torch.all(resized_landmarks.eq(expected_output))


def test_resize_landmarks_with_padding():
    # Original and new dimensions
    orig_dim = (40, 20)
    new_dim = (10, 10)
    padding = (0, 10)

    # Create a tensor of landmarks
    landmarks = torch.Tensor([[10, 20], [30, 10], [20, 5]])

    # Call the function
    resized_landmarks = resize_landmarks(landmarks, orig_dim, new_dim, padding)

    # Check that the output has the expected values
    expected_output = torch.Tensor([[2.5, 30.0 / 4], [30.0 / 4, 5], [5, 15 / 4]])
    assert torch.allclose(resized_landmarks, expected_output)

    # Create a multi-instance tensor of landmarks
    landmarks = torch.Tensor([[[10, 20], [30, 10], [20, 5]],
                              [[20, 10], [10, 30], [5, 20]]])

    # Call the function
    resized_landmarks = resize_landmarks(landmarks, orig_dim, new_dim, padding)

    # Check that the output has the expected values
    expected_output = torch.Tensor([[[2.5, 30.0 / 4], [30.0 / 4, 5], [5, 15 / 4]],
                                    [[5, 5], [2.5, 10], [5 / 4, 30 / 4]]])
    assert torch.allclose(resized_landmarks, expected_output)


def test_affine_landmarks_2d():
    # Create a tensor of 2D landmarks
    landmarks = torch.tensor([[1, 2], [3, 4], [5, 6]])

    # Create an affine matrix
    affine_matrix = torch.eye(4)

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = landmarks
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (shift by 1)
    landmarks = torch.tensor([[1, 2], [3, 4], [5, 6]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[1, 0, 0, 1],
                                  [0, 1, 0, 1],
                                  [0, 0, 1, 1],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[2, 3], [4, 5], [6, 7]])
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (90 degrees rotation)
    landmarks = torch.tensor([[1, 2], [3, 4], [5, 6]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[0, -1, 0, 0],
                                  [1, 0, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[-2, 1], [-4, 3], [-6, 5]])
    assert torch.all(transformed_landmarks.eq(expected_output))


def test_affine_landmarks_2d_multi_instance():
    # Create a multi-instance tensor of 2D landmarks
    landmarks = torch.tensor([[[1, 2], [3, 4], [5, 6]],
                              [[7, 8], [9, 10], [11, 12]]])

    # Create an affine matrix
    affine_matrix = torch.eye(4)

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = landmarks
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (shift by 1)
    landmarks = torch.tensor([[[1, 2], [3, 4], [5, 6]],
                              [[7, 8], [9, 10], [11, 12]]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[1, 0, 0, 1],
                                  [0, 1, 0, 1],
                                  [0, 0, 1, 1],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[[2, 3], [4, 5], [6, 7]],
                                    [[8, 9], [10, 11], [12, 13]]])
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (90 degrees rotation)
    landmarks = torch.tensor([[[1, 2], [3, 4], [5, 6]],
                              [[7, 8], [9, 10], [11, 12]]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[0, -1, 0, 0],
                                  [1, 0, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[[[-2, 1], [-4, 3], [-6, 5]],
                                    [[-8, 7], [-10, 9], [-12, 11]]]])
    assert torch.all(transformed_landmarks.eq(expected_output))


def test_affine_landmarks_3d():
    # Create a tensor of 3D landmarks
    landmarks = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    # Create an affine matrix
    affine_matrix = torch.eye(4)

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = landmarks
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (shift by 1)
    landmarks = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[1, 0, 0, 1],
                                  [0, 1, 0, 1],
                                  [0, 0, 1, 1],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[2, 3, 4], [5, 6, 7], [8, 9, 10]])
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (90 degrees rotation)
    landmarks = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[0, -1, 0, 0],
                                  [1, 0, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[-2, 1, 3], [-5, 4, 6], [-8, 7, 9]])
    assert torch.all(transformed_landmarks.eq(expected_output))


def test_affine_landmarks_3d_multi_instance():
    # Create a multi-instance tensor of 3D landmarks
    landmarks = torch.tensor([[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                              [[10, 11, 12], [13, 14, 15], [16, 17, 18]]])

    # Create an affine matrix
    affine_matrix = torch.eye(4)

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = landmarks
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (shift by 1)
    landmarks = torch.tensor([[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                              [[10, 11, 12], [13, 14, 15], [16, 17, 18]]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[1, 0, 0, 1],
                                  [0, 1, 0, 1],
                                  [0, 0, 1, 1],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[[2, 3, 4], [5, 6, 7], [8, 9, 10]],
                                    [[11, 12, 13], [14, 15, 16], [17, 18, 19]]])
    assert torch.all(transformed_landmarks.eq(expected_output))

    # Test with a non-identity affine matrix (90 degrees rotation)
    landmarks = torch.tensor([[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                              [[10, 11, 12], [13, 14, 15], [16, 17, 18]]])

    # Create an affine matrix
    affine_matrix = torch.Tensor([[0, -1, 0, 0],
                                  [1, 0, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 1]])

    # Call the function
    transformed_landmarks = affine_landmarks(landmarks, affine_matrix)

    # Check that the output has the expected values
    expected_output = torch.tensor([[[-2, 1, 3], [-5, 4, 6], [-8, 7, 9]],
                                    [[-11, 10, 12], [-14, 13, 15], [-17, 16, 18]]])
    assert torch.all(transformed_landmarks.eq(expected_output))


def test_affine_landmarks_invalid():
    # Create a tensor of landmarks with invalid shape
    landmarks = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8]])

    # Create an affine matrix
    affine_matrix = torch.eye(4)

    # Call the function and check that it raises a ValueError
    with pytest.raises(ValueError):
        affine_landmarks(landmarks, affine_matrix)



"""CoordConv implementation proposed in "An intriguing failing of convolutional neural networks and
    the CoordConv solution" - Liu et al."""

import torch
from torch import nn
from monai.networks.blocks import ResidualUnit


class CoordConvLayer(nn.Module):
    """
    CoordConv is a convolutional layer that adds the x and y coordinates of each pixel as additional
    channels to the input tensor. Optionally, it can also add the radial distance of each pixel to
    the center of the image as an additional channel. This is done to provide the network with
    spatial information.
        source: "An intriguing failing of convolutional neural networks and the CoordConv
            solution" - Liu et al.

    Args:
        spatial_dims (int): number of spatial dimensions of the input image.
        in_channels (int): number of input channels.
        out_channels (int): number of output channels.
        radial_channel (bool, optional): whether to add the radial distance of each pixel to the
            center of the image as an additional channel. Defaults to False.
        conv_block (nn.Module, optional): convolutional block to use. Defaults to ResidualUnit.
    """

    def __init__(self, spatial_dims: int, in_channels: int, out_channels: int,
                 radial_channel: bool = False,
                 conv_block: nn.Module = ResidualUnit  # type: ignore[assignment]
                 ):
        super().__init__()
        self.radial_channel = radial_channel
        self.add_coord_channels = AddCoordChannels(radial_channel=radial_channel)
        self.conv_block = conv_block(spatial_dims=spatial_dims,
                                     in_channels=in_channels + 2 + int(radial_channel),
                                     out_channels=out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *spatial_dims)

        Returns:
            output tensor of shape (batch_size, out_channels, *spatial_dims)
        """
        out = self.add_coord_channels(x)
        out = self.conv_block(out)
        return out


class AddCoordChannels(nn.Module):
    """
    Adds the x and y coordinates of each pixel as additional channels to the input tensor.
    Optionally, it can also add the radial distance of each pixel to the center of the image as an
    additional channel. This is done to provide the network with spatial information.

    Args:
        radial_channel (bool, optional): whether to add the radial distance of each pixel to the
            center of the image as an additional channel. Defaults to False.
    """

    def __init__(self, radial_channel: bool = False) -> None:
        super().__init__()
        self.radial_channel = radial_channel

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input tensor of shape (batch_size, in_channels, *spatial_dims)

        Returns:
            output tensor of shape (batch_size, in_channels + 2 + radial_channel, *spatial_dims)
        """
        b, _, h, w = x.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=True).to(x)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=True).to(x)
        xs, ys = torch.meshgrid(xs, ys, indexing='xy')
        xs = xs.unsqueeze(0).unsqueeze(0)
        ys = ys.unsqueeze(0).unsqueeze(0)
        xs = xs / (w - 1)
        ys = ys / (h - 1)
        xs = xs * 2 - 1
        ys = ys * 2 - 1
        xs = xs.repeat(b, 1, 1, 1)
        ys = ys.repeat(b, 1, 1, 1)
        if self.radial_channel:
            r = torch.sqrt(torch.pow(xs, 2) + torch.pow(ys, 2))
            return torch.cat((x, ys, xs, r), 1)
        return torch.cat((x, ys, xs), 1)



"""
Image transforms.
"""

from collections.abc import Mapping
from typing import Hashable

import monai
import monai.transforms
import torch
import torchvision.transforms as T  # type: ignore
from monai.transforms import Resize, SpatialPad


def resize_with_pad(
    images: torch.Tensor, dim: tuple[int, ...]
) -> tuple[torch.Tensor, tuple[int, ...]]:
    """
    Resize images to ``dim`` and pad them to preserve the aspect ratio.
        source: https://github.com/pytorch/vision/issues/6236

    Args:
        images (torch.Tensor): images to resize and pad.
        dim (tuple[int, ...]): dimension of the images.

    Returns:
        images (torch.Tensor): resized and padded images.
        padding (tuple[int, int]): padding applied to the images.
    """
    if len(dim) == 2:
        return resize_with_pad_2d(images, dim)
    elif len(dim) == 3:
        return resize_with_pad_3d(images, dim)
    raise ValueError("spatial_dims must be 2 or 3.")


def resize_with_pad_2d(
    images: torch.Tensor, dim: tuple[int, ...]
) -> tuple[torch.Tensor, tuple[int, int]]:
    """
    Resize 2D images to ``dim`` and pad them to preserve the aspect ratio.
        source: https://github.com/pytorch/vision/issues/6236

    Args:
        images (torch.Tensor): images to resize and pad.
        dim (tuple[int, int]): dimension of the images.

    Returns:
        images (torch.Tensor): resized and padded images.
        padding (tuple[int, int]): padding applied to the images.
    """
    h_orig, w_orig = images.shape[-2], images.shape[-1]
    h, w = dim[0], dim[1]
    ratio_in = h_orig / w_orig
    ratio_out = h / w

    # check if the original and final aspect ratios are the same within a margin
    if round(ratio_in, 2) != round(ratio_out, 2):
        # padding to preserve aspect ratio
        w_padding = int(h_orig / ratio_out - w_orig)
        h_padding = int(ratio_out * w_orig - h_orig)
        if w_padding > 0 and h_padding < 0:
            w_padding = w_padding // 2
            images = T.functional.pad(images, (w_padding, 0), 0, "constant")
            return T.functional.resize(images, (h, w)), (0, w_padding)

        if w_padding < 0 and h_padding > 0:
            h_padding = h_padding // 2
            images = T.functional.pad(images, (0, h_padding), 0, "constant")
            return T.functional.resize(images, (h, w)), (h_padding, 0)
    return T.functional.resize(images, (h, w)), (0, 0)


def resize_with_pad_3d(
    images: torch.Tensor, dim: tuple[int, ...]
) -> tuple[torch.Tensor, tuple[int, ...]]:
    """
    Resize 3D images to ``dim`` and pad them to preserve the aspect ratio.

    Args:
        images (torch.Tensor): images to resize and pad.
        dim (tuple[int, int, int]): dimension of the images (D, H, W).

    Returns:
        images (torch.Tensor): resized and padded images.
        padding (tuple[int, int, int]): padding applied to the images (d_pad, h_pad, w_pad).
    """
    d_orig, h_orig, w_orig = images.shape[-3], images.shape[-2], images.shape[-1]
    d, h, w = dim[0], dim[1], dim[2]

    # Calculate aspect ratios for depth-height and depth-width
    ratio_dh_in = d_orig / h_orig
    ratio_dw_in = d_orig / w_orig
    ratio_hw_in = h_orig / w_orig

    ratio_dh_out = d / h
    ratio_dw_out = d / w
    ratio_hw_out = h / w

    padding = [0, 0, 0]  # [d_padding, h_padding, w_padding]
    padded_images = images

    # Check and adjust depth-height ratio
    if round(ratio_dh_in, 2) != round(ratio_dh_out, 2):
        h_padding = int(d_orig / ratio_dh_out - h_orig)
        d_padding = int(ratio_dh_out * h_orig - d_orig)

        if h_padding > 0 and d_padding < 0:
            h_padding = h_padding // 2
            padding[1] = h_padding
        elif h_padding < 0 and d_padding > 0:
            d_padding = d_padding // 2
            padding[0] = d_padding

    # Check and adjust depth-width ratio
    if round(ratio_dw_in, 2) != round(ratio_dw_out, 2):
        w_padding = int(d_orig / ratio_dw_out - w_orig)
        d_padding = int(ratio_dw_out * w_orig - d_orig)

        if w_padding > 0 and d_padding < 0:
            w_padding = w_padding // 2
            padding[2] = w_padding
        elif (
            w_padding < 0 and d_padding > 0 and padding[0] == 0
        ):  # Only pad depth if not already padded
            d_padding = d_padding // 2
            padding[0] = d_padding

    # Check and adjust height-width ratio if needed and not already adjusted
    if round(ratio_hw_in, 2) != round(ratio_hw_out, 2) and padding[1] == 0 and padding[2] == 0:
        w_padding = int(h_orig / ratio_hw_out - w_orig)
        h_padding = int(ratio_hw_out * w_orig - h_orig)

        if w_padding > 0 and h_padding < 0:
            w_padding = w_padding // 2
            padding[2] = w_padding
        elif w_padding < 0 and h_padding > 0:
            h_padding = h_padding // 2
            padding[1] = h_padding

    # Final resize to target dimensions
    padded_images = SpatialPad((d_orig + padding[0], h_orig + padding[1], w_orig + padding[2]))(
        padded_images
    )
    resized_images = Resize((d, h, w))(padded_images)
    return resized_images, tuple(padding)


class UseOnlyFirstChannel(monai.transforms.MapTransform):
    """
    Select only the first channel of the image.
    """

    def __init__(self, keys: monai.config.KeysCollection, allow_missing_keys: bool = False) -> None:
        super().__init__(keys, allow_missing_keys)

    def __call__(
        self, data: Mapping[Hashable, monai.config.type_definitions.NdarrayOrTensor]
    ) -> dict[Hashable, monai.config.type_definitions.NdarrayOrTensor]:
        d = dict(data)
        for key in self.key_iterator(d):
            d[key] = d[key][0].unsqueeze(0)
        return d
